@-------------------------------------------------------------------------
@ Return of the Very Tiny Language for ARM
@ file : vtllib.s
@ 2003/11/07
@ 2009/03/15
@ Copyright (C) 2003-2009 Jun Mizutani <mizutani.jun@nifty.ne.jp>
@ vtllib.s may be copied under the terms of the GNU General Public License.
@-------------------------------------------------------------------------

.ifndef __VTLLIB
__VTLLIB = 1

.include "syscalls.s"
.include "signal.s"
.include "stdio.s"
.include "syserror.s"

@==============================================================
.text

MAXLINE      = 128         @ Maximum Line Length
MAX_FILE     = 256         @ Maximum Filename
MAXHISTORY   =  16         @ No. of history buffer

TIOCGWINSZ   = 0x5413

NCCS  = 19

@  c_cc characters
VTIME     = 5
VMIN      = 6

@  c_lflag bits
ISIG      = 0000001
ICANON    = 0000002
XCASE     = 0000004
ECHO      = 0000010
ECHOE     = 0000020
ECHOK     = 0000040
ECHONL    = 0000100
NOFLSH    = 0000200
TOSTOP    = 0000400
ECHOCTL   = 0001000
ECHOPRT   = 0002000
ECHOKE    = 0004000
FLUSHO    = 0010000
PENDIN    = 0040000
IEXTEN    = 0100000

TCGETS    = 0x5401
TCSETS    = 0x5402

SEEK_SET  = 0               @ Seek from beginning of file.
SEEK_CUR  = 1               @ Seek from current position.
SEEK_END  = 2               @ Seek from end of file.

@ from include/linux/wait.h
WNOHANG   = 0x00000001
WUNTRACED = 0x00000002

@ from include/asm-i386/fcntl.h
O_RDONLY =    00
O_WRONLY =    01
O_RDWR   =    02
O_CREAT  =  0100
O_EXCL   =  0200
O_NOCTTY =  0400
O_TRUNC  = 01000

S_IFMT   = 0170000
S_IFSOCK = 0140000
S_IFLNK  = 0120000
S_IFREG  = 0100000
S_IFBLK  = 0060000
S_IFDIR  = 0040000
S_IFCHR  = 0020000
S_IFIFO  = 0010000
S_ISUID  = 0004000
S_ISGID  = 0002000
S_ISVTX  = 0001000

S_IRWXU  = 00700
S_IRUSR  = 00400
S_IWUSR  = 00200
S_IXUSR  = 00100

S_IRWXG  = 00070
S_IRGRP  = 00040
S_IWGRP  = 00020
S_IXGRP  = 00010

S_IRWXO  = 00007
S_IROTH  = 00004
S_IWOTH  = 00002
S_IXOTH  = 00001

@ from include/linux/fs.h
MS_RDONLY       =  1        @ Mount read-only
MS_NOSUID       =  2        @ Ignore suid and sgid bits
MS_NODEV        =  4        @ Disallow access to device special files
MS_NOEXEC       =  8        @ Disallow program execution
MS_SYNCHRONOUS  = 16        @ Writes are synced at once
MS_REMOUNT      = 32        @ Alter flags of a mounted FS

/*------------------------------------------------------------------------
  v1-v8 ¸뤬ŪˤϼΤ褦˻Ѥ
    v1   @ Input Buffer, ѥ쥸
    v2   @ BufferSize, FileNameBuffer(FileCompletion)
    v3   @ history string ptr, FNBPointer
    v4   @          (0..v2-1)
    v5   @ current position (0..v2-1)  memory address[v1+v5]
    v6   @ DirName
    v7   @ FNArray ե̾ؤΥݥ(FileCompletion)
    v8   @ PartialName ϥХåեؤΥݥ
------------------------------------------------------------------------*/

@-------------------------------------------------------------------------
@ Խդ(ʸդ)
@   r0:Хåե, r1:ХåեƬ
@   r0 ʸ֤
@-------------------------------------------------------------------------
        .align  2
READ_LINE2:
        ldr     r2, LINE_TOP
        ldr     r3, [r2]
        str     r3, [r2, #4]        @ FLOATING_TOP=LINE_TOP
        stmfd   sp!, {r1-r2, v1-v5, lr}
        mov     ip, r0              @ Хåե
        mov     r2, r1              @ ϥХåեƬ
        mov     r0, r1              @ ϥХåեɽ
        bl      OutAsciiZ
        bl      StrLen              @ <r0:ɥ쥹, >r1:ʸ
        mov     v4, r1              @ 
        mov     r1, r2              @ ХåեƬ
        mov     r0, ip              @ Хåե
        b       RL_0

@-------------------------------------------------------------------------
@ Խդ
@   r0:Хåե, r1:ХåեƬ
@   r0 ʸ֤
@   ֤ƹƬ¸, ʣԤˤ錄ڡԲ
@-------------------------------------------------------------------------
READ_LINE3:
        stmfd   sp!, {lr}
        bl      get_cursor_position
        ldmfd   sp!, {lr}
        b       RL

@-------------------------------------------------------------------------
@ Խդ
@   r0:Хåե, r1:ХåեƬ
@   r0 ʸ֤
@-------------------------------------------------------------------------
READ_LINE:
        ldr     r2, LINE_TOP
        ldr     r3, [r2]
        str     r3, [r2, #4]        @ FLOATING_TOP=LINE=TOP
RL:     stmfd   sp!, {r1-r2, v1-v5, lr}
        mov     v4, #0              @ 
RL_0:
        ldr     v3, HistLine        @ history string ptr
        mov     v1, r1              @ Input Buffer
        mov     r2, #0
        strb    r2, [v1, v4]        @ mark EOL
        mov     v2, r0              @ BufferSize
        mov     v5, v4              @ current position
RL_next_char:
        bl      InChar
        cmp     r0, #0x1B           @ ESC ?
        bne     1f
        bl      translate_key_seq
    1:  cmp     r0, #0x09           @ TAB ?
        beq     RL_tab
        cmp     r0, #127            @ BS (linux console) ?
        beq     RL_bs
        cmp     r0, #0x08           @ BS ?
        beq     RL_bs
        cmp     r0, #0x04           @ ^D ?
        beq     RL_delete
        cmp     r0, #0x02           @ ^B
        beq     RL_cursor_left
        cmp     r0, #0x06           @ ^F
        beq     RL_cursor_right
        cmp     r0, #0x0E           @ ^N
        beq     RL_forward
        cmp     r0, #0x10           @ ^P
        beq     RL_backward
        cmp     r0, #0x0A           @ enter ?
        beq     RL_in_exit
        cmp     r0, #0x20
        blo     RL_next_char        @ illegal chars
RL_in_printable:
        add     v4, v4, #1          @ eol
        add     v5, v5, #1          @ current position
        cmp     v4, v2              @ buffer size
        bhs     RL_in_toolong
        cmp     v5, v4              @ at eol?
        blo     RL_insert           @  No. Insert Char
        bl      OutChar             @  Yes. Display Char
        sub     ip, v5, #1          @ before cursor
        strb    r0, [v1, +ip]
        b       RL_next_char
RL_insert:
        bl      OutChar
        sub     ip, v4, #1          @ p = eol-1
    1:
        cmp     v5, ip              @ while(p=>cp){buf[p]=buf[p-1]; p--}
        bhi     2f                  @   if(v5>ip) goto2
        sub     r1, ip, #1          @   r1=ip-1
        ldrb    r2, [v1, +r1]
        strb    r2, [v1, +ip]
        mov     ip, r1              @ ip--
        b       1b
    2:
        sub     ip, v5, #1
        strb    r0, [v1, +ip]       @ before cursor

        bl      print_line_after_cp
        b       RL_next_char
RL_in_toolong:
        sub     v4, v4, #1
        sub     v5, v5, #1
        b       RL_next_char
RL_in_exit:
        bl      regist_history
        bl      NewLine
        mov     r0, v4              @ eax ʸ֤
        ldmfd   sp!, {r1-r2, v1-v5, pc}    @ return

@-------------------------------------------------------------------------
@ BackSpace or Delete Character
@-------------------------------------------------------------------------
RL_bs:
        tst     v5, v5              @ if cp=0 then next_char
        beq     RL_next_char
        sub     v5, v5, #1
        adr     r0, CURSOR_LEFT
        bl      OutPString
RL_delete:
        cmp     v5, v4              @ if cp < eol then del2
        beq     RL_next_char
        sub     v4, v4, #1          @ eol--
        mov     ip, v5
    1:
        cmp     ip, v4              @ while(p<eol){buf[p]=buf[p+1]@ p++}
        beq     2f
        add     r1, ip, #1
        ldrb    r0, [v1, r1]
        strb    r0, [v1, ip]
        mov     ip, r1
        b       1b
    2:
        adr     r0, DEL_AT_CURSOR
        bl      OutPString
        b       RL_next_char

@-------------------------------------------------------------------------
@ Filename Completion
@-------------------------------------------------------------------------
RL_tab:
        bl      FilenameCompletion  @ ե̾䴰
        bl      DispLine
        b       RL_next_char

@-------------------------------------------------------------------------
RL_cursor_left:
        tst     v5, v5              @ if cp = 0 then next_char
        beq     RL_next_char
        sub     v5, v5, #1
        adr     r0, CURSOR_LEFT
        bl      OutPString
        b       RL_next_char

@-------------------------------------------------------------------------
RL_cursor_right:
        cmp     v4, v5              @ if cp=eol then next_char
        beq     RL_next_char
        add     v5, v5, #1
        adr     r0, CURSOR_RIGHT
        bl      OutPString
        b       RL_next_char

@-------------------------------------------------------------------------
RL_forward:
        bl      regist_history      @ ιԤҥȥ
        mov     r0, #1
        bl      next_history
        b       RL_disp
RL_backward:
        bl      regist_history      @ ιԤҥȥ
        mov     r0, #-1
        bl      next_history
RL_disp:
        and     r0, r0, #0x0F       @ ҥȥ 0-15
        str     r0, [v3]            @ HistLine
        bl      history2input       @ ҥȥ꤫ϥХåե
        bl      DispLine
        b       RL_next_char

@-------------------------------------------------------------------------

                .align      2
CURSOR_REPORT:  .byte       4, 0x1B
                .ascii      "[6n"               @ ^[[6n
                .align      2
SAVE_CURSOR:    .byte       2, 0x1B, '7         @ ^[7
                .align      2
RESTORE_CURSOR: .byte       2, 0x1B, '8         @ ^[8
                .align      2
DEL_AT_CURSOR:  .byte       4, 0x1B
                .ascii      "[1P"               @ ^[[1P
                .align      2
CURSOR_RIGHT:   .byte       4, 0x1B
                .ascii      "[1C"               @ ^[[1C
                .align      2
CURSOR_LEFT:    .byte       4, 0x1B
                .ascii      "[1D"               @ ^[[1D
                .align      2
CURSOR_TOP:     .byte       1, 0x0D
                .align      2
CLEAR_EOL:      .byte       4, 0x1B
                .ascii      "[0K"               @ ^[[0K
                .align      2
CSI:            .byte       2, 0x1B, '[         @ ^[[
                .align      2

@ Variable
HistLine:       .long       HistLine0    @ ɽ
HistUpdate:     .long       HistUpdate0  @ 
HistTop:        .long       history0
LINE_TOP:       .long       LINE_TOP0    @ No. of prompt characters
FLOATING_TOP:   .long       FLOATING_TOP0

.data
LINE_TOP0:      .long       7            @ No. of prompt characters
FLOATING_TOP0:  .long       7            @ Save cursor position

.text
@-------------------------------------------------------------------------
@ Ƭޡ
@   r0 : Ƭޡ
@-------------------------------------------------------------------------
set_linetop:
        stmfd   sp!, {r1, lr}
        ldr     r1, LINE_TOP
        str     r0, [r1]
        ldmfd   sp!, {r1, pc}

@--------------------------------------------------------------
@ ϥХåեҥȥϿ
@   v1 : input buffer address
@   v3 : history string ptr
@   v4 : eol (length of input string)
@   r0,r1,r2,r3 : destroy
@--------------------------------------------------------------
regist_history:
        stmfd   sp!, {lr}
        mov     r0, #0
        strb    r0, [v1, v4]        @ write 0 at eol
        bl      check_history
        tst     r1, r1
        beq     1f                  @ ƱϿѤ

        ldr     r0, [v3, #+4]       @ HistUpdate
        bl      input2history
        ldr     r0, [v3, #+4]       @ HistUpdate
        add     r0, r0, #1
        and     r0, r0, #0x0F       @ 16entry
        str     r0, [v3, #+4]       @ HistUpdate
        str     r0, [v3]            @ HistLine
    1:  ldmfd   sp!, {pc}           @ return

@--------------------------------------------------------------
@ ҥȥ r0 (1ޤ-1) ʤ
@    return : r0 = next entry
@--------------------------------------------------------------
next_history:
        stmfd   sp!, {r1-r3, v1, lr}
        ldr     v1, [v3]           @ HistLine
        mov     r2, #MAXHISTORY
        mov     r3, r0
    1:  subs    r2, r2, #1         @
        blt     2f                 @ ٤ƶʤ齪λ
        add     v1, v1, r3         @ +/-1
        and     v1, v1, #0x0F      @ wrap around
        mov     r0, v1             @ Υȥ꡼
        bl      GetHistory         @ r0 = Ƭɥ쥹
        bl      StrLen             @ <r0:ɥ쥹, >r1:ʸ
        tst     r1, r1
        beq     1b                 @ Ԥʤ鼡
    2:
        mov     r0, v1             @ ȥ꡼֤
        ldmfd   sp!, {r1-r3, v1, pc}   @ return

@--------------------------------------------------------------
@ ٤ƤΥҥȥƤɽ
@--------------------------------------------------------------
disp_history:
        stmfd   sp!, {r0-r3, lr}
        ldr     r2, HistTop
        mov     r3, #0             @ no. of history lines
    1:
        ldr     r1, HistLine
        ldr     r0, [r1]
        cmp     r3, r0
        moveq   r0, #'*'
        movne   r0, #' '
        bl      OutChar

        mov     r0, r3             @ ҥȥֹ
        mov     r1, #2             @ 2
        bl      PrintRight0
        mov     r0, #' '
        bl      OutChar
        mov     r0, r2
        bl      OutAsciiZ
        bl      NewLine
        add     r2, r2, #MAXLINE   @ next history string
        add     r3, r3, #1
        cmp     r3, #MAXHISTORY
        bne     1b                 @ check next
        ldmfd   sp!, {r0-r3, pc}   @ return

@--------------------------------------------------------------
@ ٤ƤΥҥȥƤõ
@--------------------------------------------------------------
erase_history:
        stmfd   sp!, {r0-r2, lr}
        ldr     r2, HistTop
        mov     r1, #0             @ no. of history lines
        mov     r0, #0
    1:  str     r0, [r2]
        add     r2, r2, #MAXLINE   @ next history
        add     r1, r1, #1
        cmp     r1, #MAXHISTORY
        bne     1b                 @ check next
        ldmfd   sp!, {r0-r2, pc}   @ return

@--------------------------------------------------------------
@ ϥХåեƱƤΥҥȥХåե뤫å
@   v1 : input buffer address
@   v4 : eol (length of input string)
@   r1 : if found then return 0
@   r0,r2,r3 : destroy
@--------------------------------------------------------------
check_history:
        stmfd   sp!, {v1-v3, lr}
        mov     v3, v1             @ save input buffer top
        ldr     v2, HistTop
        mov     r3, #MAXHISTORY    @ no. of history lines
    1:
        mov     v1, v3             @ restore input buffer top
        mov     r2, #0             @ string top

    2:  ldrb    r0, [v1], #1       @ compare char, v1++
        ldrb    r1, [v2, +r2]
        cmp     r0, r1
        bne     3f                 @ different char
        tst     r0, r0             @ eol ?
        beq     4f                 @ found
        add     r2, r2, #1         @ next char
        b       2b

    3:  add     v2, v2, #MAXLINE   @ next history string
        subs    r3, r3, #1
        bne     1b                 @ check next

        mov     r1, #1             @ compare all, not found
        ldmfd   sp!, {v1-v3, pc}   @ return

    4:  mov     r1, #0             @ found
        ldmfd   sp!, {v1-v3, pc}   @ return

@--------------------------------------------------------------
@ ϥХåեΥǥå򥢥ɥ쥹Ѵ
@   enter  r0 : ҥȥХåեΥǥå (0..15)
@   exit   r0 : historyinput buffer top address
@--------------------------------------------------------------
GetHistory:
        stmfd   sp!, {r1, r2, lr}
        mov     r1, #MAXLINE
        ldr     r2, HistTop
        mla     r0, r1, r0, r2     @ r0=r1*r0+r2
        ldmfd   sp!, {r1, r2, pc}  @ return

@--------------------------------------------------------------
@ ϥХåեҥȥХåեإԡ
@   r0 : ҥȥХåեΥǥå (0..15)
@   v1 : input buffer
@--------------------------------------------------------------
input2history:
        stmfd   sp!, {r0-r1, v1, lr}
        mov     r1, v1
        bl      GetHistory
        mov     v1, r0
        b       1f

@--------------------------------------------------------------
@ ҥȥХåեϥХåեإԡ
@   r0 : ҥȥХåեΥǥå (0..15)
@   v1 : input buffer
@--------------------------------------------------------------
history2input:
        stmfd   sp!, {r0-r1, v1, lr}
        bl      GetHistory
        mov     r1, r0
    1:  ldrb    r0, [r1], #1
        strb    r0, [v1], #1
        cmp     r0, #0
        bne     1b
        ldmfd   sp!, {r0-r1, v1, pc}        @ return

@--------------------------------------------------------------
@  ϥХåեץץľΰ֤ɽƥϺǽ
@  entry  v1 : ϥХåեƬɥ쥹
@--------------------------------------------------------------
DispLine:
        stmfd   sp!, {lr}
        bl      LineTop                 @ Ƭ
        mov     r0, v1
        bl      OutAsciiZ               @ ϥХåեɽ
        adr     r0, CLEAR_EOL
        bl      OutPString
        mov     r0, v1
        bl      StrLen                  @ <r0:ɥ쥹, >r1:ʸ
        mov     v4, r1                  @ ʸ
        mov     v5, v4                  @ ϰֹ
        ldmfd   sp!, {pc}               @ return

@--------------------------------------------------------------
@ ֤
get_cursor_position:
        stmfd   sp!, {r0-r3, lr}
        adr     r0, CURSOR_REPORT
        bl      OutPString
        bl      InChar                  @ ֤ʸ
        cmp     r0, #0x1B               @ ^[[y@xR
        bne     1f
        bl      InChar
        cmp     r0, #'['
        bne     1f
        bl      get_decimal             @ Y
        mov     r3, r1
        bl      get_decimal             @ X
        sub     r1, r1, #1
        ldr     r0, FLOATING_TOP
        str     r1, [r0]                @ ޡ
    1:  ldmfd   sp!, {r0-r3, pc}        @ return

get_decimal:
        stmfd   sp!, {r3, lr}
        mov     r1, #0
        mov     r3, #10
        bl      InChar
        sub     r0, r0, #'0
    1:  mul     r2, r1, r3              @
        add     r1, r0, r2
        bl      InChar
        sub     r0, r0, #'0
        cmp     r0, #9
        ble     1b
        ldmfd   sp!, {r3, pc}           @ return

@--------------------------------------------------------------
@
print_line_after_cp:
        stmfd   sp!, {lr}
        adr     r0, SAVE_CURSOR
        bl      OutPString
        adr     r0, CLEAR_EOL
        bl      OutPString
        add     r0, v5, v1              @ address
        sub     r1, v4, v5              @ length
        bl      OutString
        adr     r0, RESTORE_CURSOR
        bl      OutPString
        ldmfd   sp!, {pc}               @ return

@--------------------------------------------------------------
@ Translate Function Key into ctrl-sequence
translate_key_seq:
        stmfd   sp!, {lr}
        bl      InChar
        cmp     r0, #'[
        movne   r0, #0
        ldmnefd sp!, {pc}           @ return
        bl      InChar
        cmp     r0, #'A
        moveq   r0, #'P - 0x40      @ ^P
        ldmeqfd sp!, {pc}           @ return
        cmp     r0, #'B
        moveq   r0, #'N - 0x40      @ ^N
        ldmeqfd sp!, {pc}           @ return
        cmp     r0, #'C
        moveq   r0, #'F - 0x40      @ ^F
        ldmeqfd sp!, {pc}           @ return
        cmp     r0, #'D
        moveq   r0, #'B - 0x40      @ ^B
        ldmeqfd sp!, {pc}           @ return
        cmp     r0, #'3             @ ^[[3~ (Del)
        cmpne   r0, #'4             @ ^[[4~ (End)
        ldmnefd sp!, {pc}           @ return
        bl      InChar
        cmp     r0, #'~
        moveq   r0, #4              @ ^D
        ldmfd   sp!, {pc}           @ return

@--------------------------------------------------------------
@ Ƭ˥ư(ޡ)
LineTop:
        stmfd   sp!, {r0-r2, lr}
        adr     r0, CURSOR_TOP
        bl      OutPString
        adr     r0, CURSOR_RIGHT
        ldr     r2, FLOATING_TOP        @ ޡ
        ldr     r2, [r2]
        tst     r2, r2                  @ if 0 return
        beq     2f
    1:  bl      OutPString
        subs    r2, r2, #1
        bne     1b
    2:  ldmfd   sp!, {r0-r2, pc}        @ return

@--------------------------------------------------------------
@  ե̾䴰ǽ
@  entry  v5 : ʸϤϥХåեΰ
@         v1 : ϥХåեƬɥ쥹
@--------------------------------------------------------------
FilenameCompletion:
        stmfd   sp!, {r0-r2, v1-v8, lr}
        ldr     v2, FileNameBuffer      @ FileNameBuffer
        ldr     v6, DirName
        ldr     v7, FNArray             @ ե̾ؤΥݥ
        ldr     v8, PartialName         @ ϥХåեΥݥ
        bl      ExtractFilename         @ ϥХåեѥ̾
        ldrb    r0, [v1]                @ Ƭʸ
        cmp     r0, #0                  @ ԤĹ0
        beq     1f
        bl      GetDirectoryEntry       @ ե̾򥳥ԡ
        bl      InsertFileName          @ 䴰ϥХåե
    1:
        ldmfd   sp!, {r0-r2, v1-v8, pc}

@==============================================================
                .align  2
NoCompletion:   .asciz  "<none>"
                .align  2
current_dir:    .asciz  "./"
                .align  2
DirName:        .long   DirName0
PathName:       .long   PathName0
PartialName:    .long   PartialName0
FileNameBuffer: .long   FileNameBuffer0
FNArray:        .long   FNArray0
FNBPointer:     .long   FNBPointer0
FNCount:        .long   FNCount0
dir_ent:        .long   dir_ent0        @ 256 bytes
file_stat:      .long   file_stat0      @ 64 bytes
                .align  2

@--------------------------------------------------------------
@ פե̾ʣʤɽʤ٤Ĺ䴰롣
@
@ פե̾ʤʤ顢<none>ϥХåե
@ ˰פե̾򥳥ԡ
@ ϥХåե0ɲáϤϥХåեΰ
@ 򹹿. ϥХåեʸ(v5)֤
@--------------------------------------------------------------
InsertFileName:
        stmfd   sp!, {r0-r3, v8, lr}
        tst     v4, v4                  @ FNCount ե
        adreq   r3, NoCompletion        @ <none>ϥХåե
        beq     6f                      @ פե̾ʤ
        ldr     r0, [v8]                @ ʬե̾
        bl      StrLen                  @ r1 = ʬե̾Ĺ
        cmp     v4, #1                  @ ҤȤĤ?
        ldreq   r0, [v7]
        addeq   r3, r0, r1              @ r3 = FNArray[0] + r1
        beq     6f                      @ ϥХåե˥ԡ
        bl      ListFile                @ ե뤬ʣʤɽ

        @ ʣפƤʤ٤Ĺ䴰
        @ ǽΥȥ꡼ȼӡ٤ƤΥȥ꡼פƤ
        @ Ӥʸ1Ŀʤ롣פʤʸĤä齪
        mov     r2, #1                  @ ɲä䴰Ǥʸ
    1:
        sub     v8, v4, #1              @ ե-1
        ldr     r0, [v7]                @ ǽΥե̾
        add     r3, r0, r1              @ r3 = FNArray[0] + ʬե̾Ĺ
        ldrb    r0, [r3, +r2]           @ r0 = (FNArray[0] + Ĺ + r2)
    2:
        ldr     ip, [v7, v8,LSL #2]     @ ip = &FNArray[v8]
        add     ip, ip, r1              @ ip = FNArray[v8] + Ĺ
        ldrb    ip, [ip, +r2]           @ ip = FNArray[v8] + Ĺ + r2
        cmp     r0, ip
        bne     3f                      @ ۤʤʸȯ
        subs    v8, v8, #1              @ Υե̾
        bne     2b                      @ ٤ƤΥե̾Ƿ֤

        add     r2, r2, #1              @ ɲä䴰Ǥʸ+1
        b       1b                      @ ʸ
    3:
        cmp     r2, #1                  @ ɲʸʤ
        beq     9f                      @ ʣ뤬ɲ䴰Բ

    4:
        ldrb    r0, [r3]                @ 䴰ʬ򥳥ԡ
        strb    r0, [v1, +v5]           @ ϥХåեɲ
        subs    r2, r2, #1
        beq     8f                      @ 䴰ʬԡλ
        add     r3, r3, #1              @ ʸ
        add     v5, v5, #1
        b       4b                      @

    6:
        ldrb    r0, [r3]                @ ե̾򥳥ԡ
        strb    r0, [v1, v5]            @ ϥХåեɲ
        add     r3, r3, #1              @ ʸ
        add     v5, v5, #1
        tst     r0, r0                  @ ʸ0ǽλ
        bne     6b

    9:  ldmfd   sp!, {r0-r3, v8, pc}    @ return

    8:
        mov     r0, #0                  @ 䴰λ
        strb    r0, [v1, v5]            @ ϥХåե0
        ldmfd   sp!, {r0-r3, v8, pc}    @ return

@--------------------------------------------------------------
@ ʸ󤫤ǥ쥯ȥ̾ʬե̾Ф
@ Хåե DirName(v6), PartialName(v8,ݥ)˳Ǽ
@ TAB줿ϥХåեκǸʸս
@ 󤷤ơƬޤϥڡޤ " õ
@ ƬޤϥڡθϥХåեκǸޤǤ
@ ʸϤƥѥ̾(v6)ȥե̾(v8)Хåե¸
@  entry  v5 : ʸϤϥХåեΰ
@         v1 : ϥХåեƬɥ쥹
@--------------------------------------------------------------
ExtractFilename:
        stmfd   sp!, {lr}
        add     r3, v5, v1              @ (ϺѤ߰+1)򥳥ԡ
        mov     r1, r3
        mov     r0, #0
        strb    r0, [r1]                @ ϺѤʸޡ
        mov     v3, v2                  @ FNBPointer=FileNameBuffer
        mov     v4, #0                  @ FNCount=0
    1:
                                        @ ʬѥ̾Ƭܤ
        ldrb    r0, [r1]                @ ֤
        cmp     r0, #0x20               @ ϥѥ̾ζڤ
        beq     2f                      @ ʤ鼡ν
        cmp     r0, #'"                 @ Űѥ̾ζڤ
        beq     2f                      @ ŰǤ⼡ν
        cmp     r1, v1                  @ Ƭå
        beq     3f                      @ Ƭʤ鼡ν
        sub     r1, r1, #1              @ ˸
        b       1b                      @ ⤦Ĵ٤

    2:  add     r1, r1, #1              @ ȯΤƬ
    3:
        ldrb    r0, [r1]
        cmp     r0, #0                  @ ʸ
        bne     4f
        ldmfd   sp!, {pc}               @ ʤ(Ĺ0)ʤ齪λ

    4:  sub     r3, r3, #1              @ ϺѤʸǽɥ쥹
        ldrb    r0, [r3]
        cmp     r0, #'/                 @ ǥ쥯ȥʬ
        addeq   r3, r3, #1              @ ե̾/
        beq     5f                      @ ڤȯ
        cmp     r1, r3                  @ ǥ쥯ȥʬʤ?
        bne     4b
    5:                                  @ ǥ쥯ȥ̾򥳥ԡ
        mov     r0, #0
        strb    r0, [v6]                @ ǥ쥯ȥ̾Хåե
        str     r3, [v8]                @ ʬե̾Ƭ
        subs    r2, r3, r1              @ r2=ǥ쥯ȥ̾ʸ
        beq     8f                      @ ǥ쥯ȥʬʤ

        mov     ip, v6                  @ DirName
    7:
        ldrb    r0, [r1],#1             @ ԡ
        strb    r0, [ip],#1             @ ǥ쥯ȥ̾Хåե
        subs    r2, r2, #1
        bne     7b
        mov     r0, #0
        strb    r0, [ip]                @ ʸޡ
    8:  ldmfd   sp!, {pc}               @ ꥿

@-------------------------------------------------------------------------
@ ǥ쥯ȥΥȥgetdentsǼ(1ĤȤϸ¤ʤΤ?)ơ
@ 1ĤŤĥե/ǥ쥯ȥ̾lstatȽǤ
@ ǥ쥯ȥǰפե̾ե̾Хåե˽񤭹ࡣ
@-------------------------------------------------------------------------
GetDirectoryEntry:
        stmfd   sp!, {r0-r3, v1, lr}
        ldrb    r0, [v6]                @ ǥ쥯ȥʬκǽʸ
        tst     r0, r0                  @ Ĺ 0 ?
        adreq   r0, current_dir         @ ǥ쥯ȥʬʤ
        movne   r0, v6                  @ ǥ쥯ȥ̾Хåե
        bl      fropen                  @ ǥ쥯ȥꥪץ
        bmi     4f
        mov     r3, r0                  @ fd 
    1:                                  @ ǥ쥯ȥꥨȥ
        mov     r0, r3                  @ fd 
        ldr     r1, dir_ent             @ dir_entǼƬɥ쥹
        mov     v1, r1                  @ v1 : dir_entؤΥݥ
        ldr     r2, size_dir_ent        @ dir_entǼΰ襵
        swi     #sys_getdents           @ dir_entʣ֤
        tst     r0, r0                  @ valid buffer length
        blmi    SysCallError            @ ƥॳ륨顼
        beq     4f                      @ λ
        mov     r2, r0                  @ r2 : buffer size
    2:
        mov     r1, v1                  @ v1 : dir_entؤΥݥ
        bl      GetFileStat             @ ե
        ldr     r1, file_stat
        ldrh    r0, [r1, #+8]           @ file_stat.st_mode
        and     r0, r0, #S_IFDIR        @ ǥ쥯ȥ?
        add     r1, v1, #10             @ ե̾Ƭɥ쥹
        bl      CopyFilename            @ פե̾

        @ sys_getdents֤ȥ꤬ʣξˤϼΥե
        @ 1Ĥʤ鼡Υǥ쥯ȥꥨȥ롣
        add     r1, v1, #8              @ 쥳Ĺɥ쥹
        ldrh    r0, [r1]                @ rec_len 쥳Ĺ
        subs    r2, r2, r0              @ buffer_size - rec_len
        beq     1b                      @ Υǥ쥯ȥꥨȥ
        add     v1, v1, r0              @ Υե̾γǼΰ
        mov     r1, v1
        b       2b                      @ Υե

    4:
        mov     r0, r3                  @ fd
        bl      fclose                  @ ǥ쥯ȥꥯ
        ldmfd   sp!, {r0-r3, v1, pc}    @ return

size_dir_ent:  .long size_dir_ent0

@--------------------------------------------------------------
@ DirNamedir_ent.dnamePathName
@ PathNameΥեξ֤file_stat¤Τ˼
@ entry
@   r1 : dir_entɥ쥹
@   v6 : DirName
@   DirName ˥ǥ쥯ȥ̾
@--------------------------------------------------------------
GetFileStat:
        stmfd   sp!, {r0-r3, v1, lr}
        add     r2, r1, #10             @ dir_ent.d_name + x
        ldr     r3, PathName            @ PathName¸ꥢ
        mov     ip, r3
        mov     r0, v6                  @ DirNameǥ쥯ȥ̾¸ɥ쥹
        bl      StrLen                  @ ǥ쥯ȥ̾Ĺ>r1
        tst     r1, r1
        beq     2f
    1:
        ldrb    v1, [r0], #1            @ ǥ쥯ȥ̾Υԡ
        strb    v1, [ip], #1            @ PathName˽񤭹
        subs    r1, r1, #1              @ -1ˤʤ뤿, bneԲ
        bne     1b
    2:  mov     r0, r2                  @ ե̾Ĺ
        bl      StrLen                  @ <r0:ɥ쥹, >r1:ʸ
    3:
        ldrb    v1, [r2], #1            @ ե̾Υԡ
        strb    v1, [ip], #1            @ PathName˽񤭹
        subs    r1, r1, #1
        bne     3b
        strb    r1, [ip]                @ ʸ(0)ޡ
        mov     r0, r3                  @ ѥ̾Ƭɥ쥹
        ldr     r1, file_stat           @ file_stat0Υɥ쥹
        swi     #sys_newlstat           @ եμ
        tst     r0, r0                  @ valid buffer length
        blmi    SysCallError            @ ƥॳ륨顼
        ldmfd   sp!, {r0-r3, v1, pc}    @ return

@--------------------------------------------------------------
@ ǥ쥯ȥǰפե̾ե̾Хåե
@ (FileNameBuffer)˽񤭹
@ ե̾ǥ쥯ȥ̾ʤ"/"ղä
@ entry r0 : ǥ쥯ȥե饰
@       r1 : ե̾Ƭɥ쥹
@       v8 : ʬե̾Ƭɥ쥹ǼΰؤΥݥ
@--------------------------------------------------------------
CopyFilename:
        stmfd   sp!, {r0-r3, v1, lr}
        cmp     v4, #MAX_FILE           @ v4:FNCount Ͽե
        bhs     5f
        mov     r3, r1                  @ ե̾Ƭɥ쥹
        mov     v1, r0                  @ ǥ쥯ȥե饰
        ldr     r2, [v8]                @ v8:PartialName
    1:
        ldrb    r0, [r2], #1            @ ʬե̾
        tst     r0, r0                  @ ʸ?
        beq     2f                      @ ʬե̾ϰ
        ldrb    ip, [r1], #1            @ ե̾
        cmp     r0, ip                  @ 1ʸ
        bne     5f                      @ ۤʤнλ
        b       1b                      @ ʸ

    2:  @ פե̾ǼǤ뤫å
        mov     r0, r3                  @ ե̾Ƭɥ쥹
        bl      StrLen                  @ ե̾Ĺ
        mov     r2, r1                  @ ե̾Ĺ
        add     ip, r1, #2              @ ʸ '/#0'
        add     ip, ip, v3              @ ɲûκǽ v3:FNBPointer
        cmp     ip, v7                  @ FileNameBufferľ(FNArray0)
        bhs     5f                      @ Хåե礭ʤ:λ
        @ ե̾ХåեΥե̾Ƭɥ쥹Ͽ
        str     v3, [v7, v4,LSL #2]     @ FNArray[FNCount]=ip
        add     v4, v4, #1              @ ե̾ι
    3:
        ldrb    ip, [r3], #1            @ ե̾Υԡ
        strb    ip, [v3], #1
        subs    r2, r2, #1              @ ե̾Ĺ򷫤֤
        bne     3b

        tst     v1, v1                  @ ǥ쥯ȥե饰
        movne   r0, #'/'                @ ǥ쥯ȥ̾ʤ"/"ղ
        strneb  r0, [v3], #1
        mov     r0, #0
        strb    r0, [v3], #1            @ ѥ졼(0)
    5:
        ldmfd   sp!, {r0-r3, v1, pc}    @ return

@--------------------------------------------------------------
@ ե̾Хåեɽ
@--------------------------------------------------------------
ListFile:
        stmfd   sp!, {r0-r3, lr}
        bl      NewLine
        mov     r3, #0                  @ Ŀ
    1:
        ldr     r2, [v7, r3,LSL #2]     @ FNArray + FNCount * 4
        mov     r0, r3
        mov     r1, #4                  @ 4
        bl      PrintRight              @ ֹɽ
        mov     r0, #0x20
        bl      OutChar
        mov     r0, r2
        bl      OutAsciiZ               @ ե̾ɽ
        bl      NewLine
        add     r3, r3, #1
        cmp     r3, v4
        blt     1b
    2:
        ldmfd   sp!, {r0-r3, pc}        @ return

@--------------------------------------------------------------
@ ߤ termios ¸
@--------------------------------------------------------------
GET_TERMIOS:
        stmfd   sp!, {r0-r3, lr}
        ldr     r1, old_termios
        mov     r3, r1                  @ old_termios
        bl      tcgetattr
        ldr     r2, new_termios
        mov     r1, r3                  @ old_termios
        sub     r3, r2, r1
        mov     r3, r3, LSR #2
    1:
        ldr     r0, [r1], #4
        str     r0, [r2], #4
        subs    r3, r3, #1
        bne     1b
        ldmfd   sp!, {r0-r3, pc}        @ return

@--------------------------------------------------------------
@  termios 
@ Raw⡼, ECHO ̵, ECHONL ̵
@ VTIME=0, VMIN=1 : 1Хɤ߼ޤԵ
@--------------------------------------------------------------
SET_TERMIOS:
        stmfd   sp!, {r0-r2, lr}
        ldr     r2, new_termios
        ldr     r0, [r2, #+12]          @ c_lflag
        ldr     r1, termios_mode
        and     r0, r0, r1
        orr     r0, r0, #ISIG
        str     r0, [r2, #+12]
        mov     r0, #0
        ldr     r1, new_term_c_cc
        mov     r0, #1
        strb    r0, [r1, #VMIN]
        mov     r0, #0
        strb    r0, [r1, #VTIME]
        ldr     r1, new_termios
        bl      tcsetattr
        ldmfd   sp!, {r0-r2, pc}        @ return

new_termios:    .long   new_termios0
old_termios:    .long   old_termios0
termios_mode:   .long   ~ICANON & ~ECHO & ~ECHONL
new_term_c_cc:  .long   nt_c_cc

@--------------------------------------------------------------
@ ¸Ƥ termios 
@--------------------------------------------------------------
RESTORE_TERMIOS:
        stmfd   sp!, {r0-r2, lr}
        ldr     r1, old_termios
        bl      tcsetattr
        ldmfd   sp!, {r0-r2, pc}        @ return

@--------------------------------------------------------------
@ ɸϤ termios μ
@ tcgetattr(&termios)
@ tcsetattr(&termios)
@ r0 : destroyed
@ r1 : termios buffer adress
@ r2 : destroyed
@--------------------------------------------------------------
tcgetattr:
        ldr     r0, TC_GETS
        b       IOCTL

tcsetattr:
        ldr     r0, TC_SETS

@--------------------------------------------------------------
@ ɸϤ ioctl μ¹
@ sys_ioctl(unsigned int fd, unsigned int cmd,
@           unsigned long arg)
@ r0 : cmd
@ r1 : buffer adress
@--------------------------------------------------------------
IOCTL:
        mov     r2, r1                  @ set arg
        mov     r1, r0                  @ set cmd
        mov     r0, #0                  @ 0 : to stdin
        swi     #sys_ioctl
        mov     pc, lr                  @ return

TC_GETS:    .long   TCGETS
TC_SETS:    .long   TCSETS

@--------------------------------------------------------------
@ input 1 character from stdin
@ eax : get char (0:not pressed)
@--------------------------------------------------------------
RealKey:
        stmfd   sp!, {r1-r4, lr}
        ldr     r3, new_term_c_cc
        mov     r0, #0
        strb    r0, [r3, #VMIN]
        ldr     r1, new_termios
        bl      tcsetattr
        mov     r0, #0                  @ r0  stdin
        stmfd   sp!, {r0}
        mov     r1, sp                  @ r1  address
        mov     r2, #1                  @ r2  length
        swi     #sys_read
        ldmfd   sp!, {r1}               @ pop char
        tst     r0, r0                  @ if 0 then empty
        moveq   r4, r0
        movne   r4, r1                  @ char code
        mov     r1, #1
        strb    r1, [r3, #VMIN]
        ldr     r1, new_termios
        bl      tcsetattr
        mov     r0, r4
        ldmfd   sp!, {r1-r4, pc}        @ return

@-------------------------------------------------------------------------
@ get window size
@ r0 : column(upper 16bit), raw(lower 16bit)
@-------------------------------------------------------------------------
WinSize:
        stmfd   sp!, {r1-r2, lr}
        mov     r0, #0                  @ to stdout
        ldr     r1, TIOCG_WINSZ         @ get wondow size
        ldr     r2, wsize
        swi     #sys_ioctl
        ldr     r0, [r2]                @ winsize.ws_row
        ldmfd   sp!, {r1-r2, pc}        @ return

wsize:          .long   winsize
TIOCG_WINSZ:    .long   TIOCGWINSZ

@-------------------------------------------------------------------------
@ ե򥪡ץ
@ enter   r0: 裱 filename
@ return  r0: fd, if error then r0 will be negative.
@ destroyed r1
@-------------------------------------------------------------------------
fropen:
        stmfd   sp!, {r1, r2, lr}
        mov     r1, #O_RDONLY           @ 裲 flag
        b       1f
fwopen:
        stmfd   sp!, {r1, r2, lr}
        ldr     r1, fo_mode
    1:
        mov     r2, #0644               @ 裳 mode
        swi     #sys_open               @ ƥॳֹ
        tst     r0, r0                  @ r0 <- fd
        ldmfd   sp!, {r1, r2, pc}

fo_mode:    .long   O_CREAT | O_WRONLY | O_TRUNC

@-------------------------------------------------------------------------
@ ե򥯥
@ enter   r0 : 裱 եǥץ
@-------------------------------------------------------------------------
fclose: swi     #sys_close
        mov     pc, lr

@==============================================================
.bss
                    .align  2
HistLine0:          .long   0
HistUpdate0:        .long   0
input0:             .skip   MAXLINE

                    .align 2
history0:           .skip   MAXLINE * MAXHISTORY

                    .align  2
DirName0:           .skip   MAXLINE
PathName0:          .skip   MAXLINE

                    .align  2
PartialName0:       .long   1           @ ʬե̾Ƭɥ쥹Ǽ
FileNameBuffer0:    .skip   2048, 0     @ 2kbyte for filename completion
FNArray0:           .skip   MAX_FILE*4  @ long* Filename[0..255]
FNBPointer0:        .long   1           @ FileNameBufferγǼѤߥɥ쥹+1
FNCount0:           .long   1           @ No. of Filenames

                    .align 2
old_termios0:
ot_c_iflag:         .long   1           @ input mode flags
ot_c_oflag:         .long   1           @ output mode flags
ot_c_cflag:         .long   1           @ control mode flags
ot_c_lflag:         .long   1           @ local mode flags
ot_c_line:          .byte   1           @ line discipline
ot_c_cc:            .skip   NCCS        @ control characters

                    .align 2
new_termios0:
nt_c_iflag:         .long   1           @ input mode flags
nt_c_oflag:         .long   1           @ output mode flags
nt_c_cflag:         .long   1           @ control mode flags
nt_c_lflag:         .long   1           @ local mode flags
nt_c_line:          .byte   1           @ line discipline
nt_c_cc:            .skip   NCCS        @ control characters

                    .align 2
new_sig:
nsa_sighandler:     .long   0           @  0
nsa_mask:           .long   0           @  4
nsa_flags:          .long   0           @  8
nsa_restorer:       .long   0           @ 12
old_sig:
osa_sighandler:     .long   0           @ 16
osa_mask:           .long   0           @ 20
osa_flags:          .long   0           @ 24
osa_restorer:       .long   0           @ 28

TV:
tv_sec:             .long   1
tv_usec:            .long   1
TZ:
tz_minuteswest:     .long   1
tz_dsttime:         .long   1

winsize:
ws_row:             .hword  1
ws_col:             .hword  1
ws_xpixel:          .hword  1
ws_ypixel:          .hword  1

ru0:                                @ 18 words
ru_utime_tv_sec:    .long   1       @ user time used
ru_utime_tv_usec:   .long   1       @
ru_stime_tv_sec:    .long   1       @ system time used
ru_stime_tv_usec:   .long   1       @
ru_maxrss:          .long   1       @ maximum resident set size
ru_ixrss:           .long   1       @ integral shared memory size
ru_idrss:           .long   1       @ integral unshared data size
ru_isrss:           .long   1       @ integral unshared stack size
ru_minflt:          .long   1       @ page reclaims
ru_majflt:          .long   1       @ page faults
ru_nswap:           .long   1       @ swaps
ru_inblock:         .long   1       @ block input operations
ru_oublock:         .long   1       @ block output operations
ru_msgsnd:          .long   1       @ messages sent
ru_msgrcv:          .long   1       @ messages received
ru_nsignals:        .long   1       @ signals received
ru_nvcsw:           .long   1       @ voluntary context switches
ru_nivcsw:          .long   1       @ involuntary

                    .align 2
dir_ent0:                           @ 256 bytesdir_entǼΰ
@ de_d_ino:         .long   1       @ 0
@ de_d_off:         .long   1       @ 4
@ de_d_reclen:      .hword  1       @ 8
@ de_d_name:                        @ 10    ǥ쥯ȥꥨȥ̾
                    .skip   256

                    .align  2
size_dir_ent0 = . - dir_ent0

                    .align 2
file_stat0:                         @ 64 bytes
fs_st_dev:          .hword  1       @ 0  եΥǥХֹ
fs___pad1:          .hword  1       @ 2
fs_st_ino:          .long   1       @ 4  եinodeֹ
fs_st_mode:         .hword  1       @ 8  եΥȥ
fs_st_nlink:        .hword  1       @ 10
fs_st_uid:          .hword  1       @ 12
fs_st_gid:          .hword  1       @ 14
fs_st_rdev:         .hword  1       @ 16
fs___pad2:          .hword  1       @ 18
fs_st_size:         .long   1       @ 20 ե륵(byte)
fs_st_blksize:      .long   1       @ 24 ֥å
fs_st_blocks:       .long   1
fs_st_atime:        .long   1       @ 32 եκǽ
fs___unused1:       .long   1
fs_st_mtime:        .long   1       @ 40 եκǽ
fs___unused2:       .long   1
fs_st_ctime:        .long   1       @ 48 եΤޤinodeκǽ
fs___unused3:       .long   1
fs___unused4:       .long   1
fs___unused5:       .long   1       @ 60

.endif
